from typing import Literal, List, Optional, Any
from pydantic import BaseModel, Field
from sentence_transformers.cross_encoder import CrossEncoder
from broai.interface import Context
from broai.experiments.utils import experiment

@experiment
class ReRanker(BaseModel):
    model_id:Literal["cross-encoder/ms-marco-MiniLM-L6-v2"] = Field(default="cross-encoder/ms-marco-MiniLM-L6-v2")
    model:Any = None

    def __init__(self, **data):
        super().__init__(**data)
        self.model = self.get_model()

    def get_model(self):
        return CrossEncoder(model_name_or_path=self.model_id)

    def run(self, search_query:str, contexts:List[Context], top_n:int=5):
        paired_sentences = [[search_query, context.context] for context in contexts]
        scores = self.model.predict(paired_sentences)
        top_rank = scores.argsort()[::-1][:top_n]
        ranked_contents = [contexts[n] for n in top_rank]
        ranked_scores = scores[top_rank]
        return ranked_contents, ranked_scores.tolist()


from datetime import datetime
from zoneinfo import ZoneInfo  # Available in Python 3.9+

def get_timestamp():
    dt = datetime.now(ZoneInfo("Asia/Bangkok"))
    return dt.strftime('%Y-%m-%d %H:%M:%S.%f')


import re
from typing import List
from broai.experiments.utils import experiment
from broai.interface import Context

@experiment
def split_markdown(markdown_text: str) -> List[str]:
    headings = re.findall(r'^(#+)', markdown_text, re.MULTILINE)

    # Get the maximum number of '#' found
    max_hashes = max(map(len, headings), default=0)
    print(f"Markdown headings: max({max_hashes})")
    # using % method works but f string didn't
    # retmember the pattern must not be space inside {1,&d}
    pattern = r'\n#{1,%d} ' % max_hashes
    
    result = re.split(pattern, markdown_text)
    return result

@experiment
def consolidate_markdown(chunks: List[str]) -> List[str]:
    consolidated = []
    max_len = len(chunks)
    idx = 0
    chunk_size = 10
    while idx < max_len:
        current_chunk = chunks[idx]
        if idx == max_len:
            consolidated.append(current_chunk)
            break
        elif len(current_chunk.split(" ")) < chunk_size:
            consolidated.append(current_chunk + "\n" + chunks[idx + 1])
            idx += 2
        else:
            consolidated.append(current_chunk)
            idx += 1
    return consolidated

def clean_text(text:str) -> str:
    # Remove the <span> tag and any content inside it
    result = re.sub(r'<span.*?</span>', '', text).strip()
    return result

@experiment
def get_markdown_sections(chunks:List[str]) -> List[str]:
    sections = []
    for enum, chunk in enumerate(chunks):
        # print(enum, clean_text(chunk.split("\n")[0]), len(chunk.split(" ")))
        # print(clean_text(chunk.split("\n")[0]), len(chunk.split(" ")))
        sections.append(clean_text(chunk.split("\n")[0]))
    return sections

@experiment
def split_overlap(contexts:List[Context], max_tokens:int=500, overlap:int=150) -> List[Context]:
    new_contexts:List[Context] = []
    for context in contexts:
        data_content = context.context
        tokens = data_content.split(" ")
        max_len = len(tokens)
        if max_len <= max_tokens:
            new_contexts.append(context)
        else:
            start_idx = 0
            end_idx = max_tokens
            while True:
                new_contexts.append(
                    Context(
                        context=" ".join(tokens[start_idx:end_idx]),
                        metadata=context.metadata
                    )
                )
                if end_idx > max_len:
                    break
                start_idx = end_idx - overlap
                end_idx = start_idx + max_tokens
    for idx in range(len(new_contexts)):
        metadata = new_contexts[idx].metadata.copy()
        metadata['sequence'] = idx
        new_contexts[idx].metadata = metadata
    return new_contexts

def which_section(text:str) -> str:
    if re.search(r"(abstract|abstracts)\n", text.lower()):
        return "abstract"
    if re.search(r"(introduction|background)\n", text.lower()):
        return "introduction"
    if re.search(r"(methodology|method|methods|experiment|experiments)\n", text.lower()):
        return "methodology"
    if re.search(r"(conclusion|discussion|summary)\n", text.lower()):
        return "conclusion"
    if re.search(r"(reference|references|bibliography|works cited|citations|cited works)\n", text.lower()):
        return "reference"
    return "others"

def chunk_chunks(chunks:List[str]) -> None:
    for enum, chunk in enumerate(chunks):
        tokens = chunk.split(" ")
        print(f"[{enum}] | tokens: {len(tokens)} | chars: {len(chunk)}")


from typing import Any, Dict, List, Literal
import boto3
import requests
import os

from broai.llm_management.interface import LLMChatInterface

class BedrockOllamaChat(LLMChatInterface):
    def __init__(
            self, 
            model_name:str="us.meta.llama3-2-11b-instruct-v1:0", 
            temperature:float=0, 
            region_name:str="us-west-2",
            aws_access_key_id:str=None,
            aws_secret_access_key:str=None,
            aws_session_token:str=None
    ):
        self.model_name = model_name
        self.temperature = temperature
        self.region_name = region_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token

    def get_model(self, ):
        param = {
            "service_name": "bedrock-runtime",
            "region_name": self.region_name,
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "aws_session_token": self.aws_session_token,
        }
        return boto3.client(**param)

    def UserMessage(self, text:str)->Dict[str, str]:
        return {"role": "user", "content": [{"text": text}]}
    
    def AIMessage(self, text:str)->Dict[str, str]:
        return {"role": "assistant", "content": [{"text": text}]}
    
    def SystemMessage(self, text:str)->List[Dict[str, str]]:
        return [{"text": text}]

    def run(self, system_prompt:str, messages:List[Dict[str, str]])->str:
        model = self.get_model()
        response = model.converse(
            modelId=self.model_name,
            messages=messages,
            system=self.SystemMessage(system_prompt),
            inferenceConfig={"temperature": self.temperature},
        )
        return response['output']['message']['content'][0]['text']



from abc import ABC, abstractmethod
from typing import Any

class LLMChatInterface(ABC):

    @abstractmethod
    def UserMessage(self, text:str) -> Any:
        pass

    @abstractmethod
    def AIMessage(self, text:str) -> Any:
        pass

    @abstractmethod
    def SystemMessage(self, text:str) -> Any:
        pass

    @abstractmethod
    def run(self, formatted_prompt: Any) -> str:
        pass


from typing import Optional, List, Dict, Any

def get_batch_update_query(table:str, schemas:dict, data:str, ref_keys:list)->str:
    query = f"""
    CREATE TEMP TABLE updates ({", ".join([f"{field} {schema}" for field, schema in schemas.items()])});
    INSERT INTO updates ({", ".join([f"{field}" for field in schemas.keys()])}) VALUES {data};

    UPDATE {table}
    SET {", ".join([f"{field} = updates.{field}" for field in schemas.keys() if field not in ref_keys])}
    FROM updates
    WHERE {" AND ".join([f"{table}.{field} = updates.{field}" for field in schemas.keys() if field in ref_keys])};
    """.strip()
    return query

def get_create_table_query(table:str, schemas:Dict[str, Any])->str:
    param = ",\n\t".join([f"{field} {dtype}".replace('', '') for field, dtype in schemas.items()])
    query = f"""
    CREATE TABLE IF NOT EXISTS {table} (\n\t{param}\n);
    """.strip()
    return query

def get_insert_query(table:str, fields:List[str], data:str)->str:
    query = f"""
    INSERT INTO {table} ({", ".join(fields).replace(";", "")}) VALUES {data.replace(";", "")};
    """.strip()
    return query

def get_select_query(table:str, fields:List[str], where:Optional[str]=None)->str:
    query = f"""
    SELECT {", ".join(fields)} FROM {table} {where};
    """
    return query

def get_delete_query(table:str, where_condition:str):
    where_condition = where_condition.replace(";", "").strip()
    text = "where "
    if where_condition.lower().startswith(text):
        where_condition = where_condition[len(text):]
    query = f"""DELETE FROM {table} WHERE {where_condition};"""
    return query

def replace_single_quote(text:str):
    return text.replace("'", "<|single_quote|>")

def reverse_single_quote(text:str):
    return text.replace("<|single_quote|>", "'")

class DataTypeConversion:
    @staticmethod
    def convert_single_quote(text:str):
        return text.replace("'", "<|single_quote|>")

    @staticmethod        
    def reverse_single_quote(text:str):
        return text.replace("<|single_quote|>", "'")


from typing import Any, List, Optional, Union
from pydantic import BaseModel, Field, field_validator
from broai.prompt_management.utils import get_input, get_return_json_schema, get_json_schema, parse_json_output
from broai.prompt_management.interface import Persona, Instructions, Example, Examples

class PromptGenerator(BaseModel):
        persona:Union[str, Persona] = None
        instructions:Union[str, Instructions] = None
        structured_output:Any = None
        examples:Union[str, Examples] = None
        fallback:Any = None

        @field_validator("structured_output")
        @classmethod
        def validate_structured_output(cls, v):
            if isinstance(v, str):
                return v
            try:
                issubclass(v, BaseModel)
                return v
            except Exception as e:
                raise TypeError(f"structured_output must be: str, BaseModel")
    
        def get_prompt(self, obj, interface):
            # if the input is of type: str, None
            if (obj is None) or isinstance(obj, str):
                return obj
            # if the input is of types: Persona, Instructions, Examples
            if isinstance(obj, interface):
                return obj.as_prompt()
            # if the input isn't of type: BaseModel
            if not isinstance(obj, interface):
                return get_json_schema(obj)
            raise TypeError(f"{type(obj)} is invalid type, try None, {str}, {BaseModel} instead")
    
        def as_prompt(self):
            persona = self.get_prompt(self.persona, Persona)
            instructions = self.get_prompt(self.instructions, Instructions)
            structured_output = self.get_prompt(self.structured_output, BaseModel)
            examples = self.get_prompt(self.examples, Examples)
            prompt = []
            for p in [persona, instructions, structured_output, examples]:
                if p:
                    prompt.append(p)
            return "\n\n".join(prompt)


import re
from pydantic import BaseModel
import json
from typing import List, Dict, Any

def get_input(pydantic_obj:BaseModel):
    prompt = []
    for field_name, content in pydantic_obj.model_dump().items():
        prompt.append(f"{field_name.capitalize()}: \n{content}")
    return "".join(prompt)

_get_json_schema_prompt="""\
IMPORTANT:
	- Output only in a specified JSON format.
	- Do not include any premable, postmable, introductory, conversation, explanation or additional information.

Understand the following JSON schema and output only in a specified JSON format as in examples: 
""".strip()

def get_json_schema(pydantic_obj:BaseModel, json_instructions: str = _get_json_schema_prompt):
    prompt = f"""{json_instructions}\n{json.dumps(pydantic_obj.model_json_schema(), indent=4)}"""
    return prompt

def get_return_json_schema(object:BaseModel):
    prompt = f"""Output:\n```json\n{json.dumps(object.model_dump(), indent=4)}\n```"""
    return prompt

def parse_json_output(text:str):
    text = text.split("```")
    if len(text) == 1:
        text = text[0]
    else:
        text = "".join(text[1:])
    if text.startswith("json"):
        text = text[4:]

    text = text.split("```")
    if len(text) == 1:
        text = text[0]
    else:
        text = text[1]
    return text


import functools
import warnings

class experiment:
    default_message = "You're using an experimental module, which is subject to change in future."

    def __init__(self, message=None):
        self.message = message or self.default_message

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f"[EXPERIMENT] {self.message}: {func.__name__}",
                category=UserWarning,
                stacklevel=2
            )
            return func(*args, **kwargs)
        return wrapper

    def __class_getitem__(cls, _):  # For typing support (optional)
        return cls

    @classmethod
    def __class_call__(cls, func):
        # Used when called like @experiment with no parentheses
        return cls()(func)

    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and callable(args[0]) and not kwargs:
            # Handles: @experiment with no ()
            return cls().__call__(args[0])
        return super().__new__(cls)

    @classmethod
    def message(cls, custom_message):
        return cls(message=custom_message)



import duckdb
from typing import Any, Literal, List, Dict, Optional, Tuple
from broai.duckdb_management.utils import get_create_table_query, get_select_query, get_insert_query, get_batch_update_query, get_delete_query
import pandas as pd
import os

class DuckStoreInterface:
    def __init__(self, db_name:str, table:str, schemas:dict):
        self.db_name = db_name
        self.table = table
        self.__schemas = schemas

    def show_schemas(self)->Dict[str, Any]:
        return self.__schemas
    
    def sql(self, query:str)->None:
        with duckdb.connect(self.db_name) as conn:
            conn.sql(query)
    
    def sql_df(self, query:str)->pd.DataFrame:
        with duckdb.connect(self.db_name) as conn:
            return conn.sql(query).df()

    def executemany(self, query, rows:List[Tuple[Any]]):
        with duckdb.connect(self.db_name) as con:
            con.executemany(query, rows)
    
    def delete_table(self)->None:
        query = f"""DELETE FROM {self.table};"""
        self.sql(query)

    def drop_table(self)->None:
        query = f"""DROP TABLE {self.table};"""
        self.sql(query)

    def create_table(self)->None:
        query = get_create_table_query(table=self.table, schemas=self.show_schemas())
        self.sql(query)

    def remove_database(self, confirm:str=None)->None:
        if confirm == f"remove {self.db_name}":
            os.remove(self.db_name)
            return
        print(f"If you want to remove database, use confirm 'remove {self.db_name}'")

    def add(self, fields:List[str], data:str):
        query = get_insert_query(table=self.table, fields=fields, data=data)
        self.sql(query)

    def read(self, fields:List[str]=None, where:Optional[str]=None)->None:
        if fields is None:
            fields = ["*"]
        query = get_select_query(table=self.table, fields=fields, where=where)
        return self.sql_df(query)

    def update(self, schemas:Dict[str, Any], data:str, ref_keys:List[str])->None:
        query = get_batch_update_query(table=self.table, schemas=schemas, data=data, ref_keys=ref_keys)
        self.sql(query)

    def delete(self, where_condition:str)->None:
        query = get_delete_query(table=self.table, where_condition=where_condition)
        self.sql(query)